/*
 * Segment 2 (KEx),  SWE2.4 Email/Fax/SMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.emailfaxsms;

import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.ConcurrentLinkedQueue;

import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import gnu.io.SerialPort;

/**
 * Liest asynchron aus der seriellen Schnittstelle in einen Ringpuffer.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 */
public class AsyncSerialPortReader extends Thread {

	/**
	 * Der Logger.
	 */
	protected static final Debug LOGGER = Debug.getLogger();

	/**
	 * InputStream der seriellen Schnittstelle.
	 */
	private final InputStream is;
	/**
	 * True, wenn man den Worker-Thread beenden soll.
	 */
	protected boolean beenden = false;

	/**
	 * Wird zum Debug-zwecken benutzt.
	 */
	protected boolean mitDebug = false;

	/**
	 * Der RingPuffer.
	 */
	protected byte[] puffer;
	/**
	 * Index der Nutzdaten Ende im RingPuffer.
	 */
	protected int inpEnde = 0;

	/**
	 * Index der Nutzdaten Start im RingPuffer.
	 */
	protected int inpStart = 0;

	/**
	 * Enth&auml;lt die geparsten AT-Kommando Antworten.
	 */
	protected ConcurrentLinkedQueue<String> antworte = new ConcurrentLinkedQueue<String>();

	/**
	 * Entscheidet ob der Puffer voll oder leer ist, intStart start == inpEnde.
	 */
	protected boolean voll = false;
	/**
	 * Ein Objekt das nur als mutex benutzt wird.
	 */
	protected Object mutex = new Object();
	/**
	 * Beim Datenversand wird Reader gestoppt, damit der Writer die
	 * Datenflusssteuerung auslesen kann.
	 */
	protected boolean stopped = false;
	/**
	 * Die String-Namen der ersten 33 ASCII-Charakter, die sonst nicht gedruckt
	 * werden koennen.
	 */
	protected String[] asciiCmd = new String[] { "<NUL>", "<SOH>", "<STX>", "<ETX>", "<EOT>", "<ENQ>", "<ACK>", "<BEL>",
			"<BS>", "<HT>", "<LF>", "<VT>", "<FF>", "<CR>", "<SO>", "<SI>", "<DLE>", "<DC1/XON>", "<DC2>", "<DC3/X0FF>",
			"<DC4>", "<NAK>", "<SYN>", "<ETB>", "<CAN>", "<EM>", "<SUB>", "<ESC>", "<FS>", "<GS>", "<RS>", "<US>",
			"<SP>" };

	/**
	 * Ready-Flag.
	 */
	private boolean ready = false;

	/**
	 * StandardKonstruktor.
	 *
	 * @param port
	 *            Die serielle Schnittstelle
	 * @param pufferSize
	 *            PufferGroesse
	 * @throws IOException
	 *             bei Ausnahmen
	 */
	public AsyncSerialPortReader(final SerialPort port, final int pufferSize) throws IOException {
		this.is = port.getInputStream();
		this.puffer = new byte[pufferSize];
	}

	/**
	 * Worker-Thread beenden.
	 */
	public void beenden() {
		synchronized (mutex) {
			this.beenden = true;
			try {
				this.is.close();
			} catch (final IOException e) {
				LogTools.log(AsyncSerialPortReader.LOGGER, EmailFaxSmsMsg.WarningDatenLesen, e.getMessage());
			}
			this.mutex.notify();
		}
	}

	/**
	 * Wird bei Ankunft neuer Daten gerufen.
	 */
	public void datenEingekommen() {
		synchronized (mutex) {
			this.mutex.notify();
		}
	}

	/**
	 * Ergibt den stopped Paramterer.
	 *
	 * @return stopped
	 */
	public synchronized boolean getStoppend() {
		return stopped;
	}

	/**
	 * Parst AT-Kommandos aus dem Ringpuffer aus.
	 */
	public synchronized void parseATKommandos() {
		int i, j;
		for (;;) {
			// Wenn inpEnde == inpStart und Puffer nicht voll ist, dann ist es
			// leer
			if ((inpEnde == inpStart) && !voll) {
				return;
			}
			i = inpStart;
			// Wenn voll ist dann i == inpEnde, obwohl wir noch nicht am Ende
			// sind
			if (voll && ((puffer[i] == AsyncSerialPort.CR) || (puffer[i] == AsyncSerialPort.LF))) {
				i = ++i % puffer.length;
			}
			// Beginn der AT Antwort
			while (((puffer[i] == AsyncSerialPort.CR) || (puffer[i] == AsyncSerialPort.LF)) && (i != inpEnde)) {
				i = ++i % puffer.length;
			}
			if (i == inpEnde) {
				return;
			}
			j = (i + 1) % puffer.length;
			// Ende der AT Antwort
			while (((puffer[j] != AsyncSerialPort.CR) && (puffer[j] != AsyncSerialPort.LF)) && (j != inpEnde)) {
				j = ++j % puffer.length;
			}
			if (j == inpEnde) {
				return;
			}

			final StringBuffer strBuff = new StringBuffer((i < j) ? j - i : puffer.length - (i - j));
			while (i != j) {
				final byte b = puffer[i++];
				i = i % puffer.length;
				if ((b < asciiCmd.length) && (b >= 0)) {
					strBuff.append(asciiCmd[b]);
				} else {
					strBuff.append((char) b);
				}
			}
			// wegen binaeren kommuniation wollen wir die CR LF charakter
			// wegwerfen
			while (((puffer[j] == AsyncSerialPort.CR) || (puffer[j] == AsyncSerialPort.LF)) && (j != inpEnde)) {
				j = ++j % puffer.length;
			}
			LogTools.log(Debug.getLogger(), EmailFaxSmsMsg.FinestModemAntwort, strBuff.toString());
			antworte.add(strBuff.toString());
			inpStart = j;
			voll = false;
		}
	}

	/**
	 * Gibt zurück, ob der Thread bereit ist.
	 *
	 * @return bereit oder nicht
	 */
	public boolean ready() {
		return ready;
	}

	/**
	 * Empf&auml;ngt ein AT Kommando Response.
	 *
	 * @param timeout
	 *            Timeout
	 * @return AT Response
	 */
	public synchronized String recvATResp(final long timeout) {
		long start;
		long resttime = timeout;
		parseATKommandos();
		while (antworte.isEmpty() && (resttime > 0)) {
			try {
				start = System.currentTimeMillis();
				this.wait(resttime);
				parseATKommandos();
				resttime = resttime - (System.currentTimeMillis() - start);
			} catch (final InterruptedException e) {
				LogTools.log(AsyncSerialPortReader.LOGGER, EmailFaxSmsMsg.WarningAntwortWartenAbbruch);
			}
		}
		if (antworte.isEmpty()) {
			return null;
		}
		return antworte.poll();
	}

	/**
	 * Empf&auml;ngt die erwartete ByteKette.
	 *
	 * @param erwartet
	 *            Erwartete ByteKette
	 * @param timeout
	 *            Timeout
	 * @return true, wenn erwartete bytes gekommen sind
	 */
	public synchronized boolean recvBytes(final long timeout, final byte[] erwartet) {
		long start;
		long resttime = timeout;
		int i, j, k;
		while (resttime > 0) {
			try {
				// Es kann sein es, gibt vor unserem byte noch irgendwelche
				// Kommandos
				parseATKommandos();

				// for (int g = inpStart; g < inpEnde; g++) {
				// System.out.println("BYTE: " + puffer[g]);
				// }

				// Uberspringen der CR LF Charakter
				k = inpStart;
				while (((puffer[k] == AsyncSerialPort.CR) || (puffer[k] == AsyncSerialPort.LF)) && (k != inpEnde)) {
					k = ++k % puffer.length;
				}
				// Finden wir den ByteSteam
				if (puffer[k] == erwartet[0]) {
					for (i = 0, j = k; i < erwartet.length; i++, j = ++j % puffer.length) {
						if (erwartet[i] != puffer[j]) {
							break;
						}
					}
					if (i >= erwartet.length) {
						if (mitDebug) {
							final StringBuffer strBuff = new StringBuffer(
									(inpStart < j) ? j - inpStart : puffer.length - (inpStart - j));
							while (inpStart != j) {
								final byte b = puffer[inpStart++];
								inpStart = inpStart % puffer.length;
								if ((b < asciiCmd.length) && (b >= 0)) {
									strBuff.append(asciiCmd[b]);
								} else {
									strBuff.append((char) b);
								}
							}
							LogTools.log(Debug.getLogger(), EmailFaxSmsMsg.FinestModemAntwort, strBuff.toString());
						}
						inpStart = j;
						return true;
					}
				}
				start = System.currentTimeMillis();
				this.wait(resttime);
				resttime -= System.currentTimeMillis() - start;
			} catch (final InterruptedException e) {
				LogTools.log(AsyncSerialPortReader.LOGGER, EmailFaxSmsMsg.WarningAntwortWartenAbbruch);
			}
		}
		return false;
	}

	/**
	 * Der Worker-Thread.
	 */
	@Override
	public void run() {
		while (!beenden) {
			try {
				ready = true;
				synchronized (this) {
					while (!stopped && !beenden && !voll && (is.available() > 0)) {
						final int c = is.read();
						// System.out.println("IN: " + c);
						puffer[inpEnde++] = (byte) c;
						inpEnde = inpEnde % puffer.length;
						if (inpEnde == inpStart) {
							voll = true;
						}
					}
					this.notify();
				}
				synchronized (mutex) {
					if (!beenden) {
						this.mutex.wait();
					}
				}
			} catch (final InterruptedException e) {
				LogTools.log(AsyncSerialPortReader.LOGGER, EmailFaxSmsMsg.WarningThreadUnterbrochen);
			} catch (final IOException e) {
				LogTools.log(AsyncSerialPortReader.LOGGER, EmailFaxSmsMsg.WarningLeseFehler, e.getMessage());
			}
		}
		LogTools.log(AsyncSerialPortReader.LOGGER, EmailFaxSmsMsg.FinerThreadBeendet, this.getClass().getName());
	}

	/**
	 * Setzt den Debug-Parameter.
	 *
	 * @param debug
	 *            true, wenn Debug Ein-ist
	 */
	public void setDebug(final boolean debug) {
		this.mitDebug = debug;
	}

	/**
	 * Setzt den stopped Parameter.
	 *
	 * @param stop
	 *            Stop ja/nein
	 */
	public synchronized void setStopped(final boolean stop) {
		this.stopped = stop;
	}

}
